@@ -7,7 +7,10 @@ These checks catch errors relating to the use of ``-
 
 import os
 import re
+from pathlib import Path
+from typing import Any
 
+from .. import utils as _utils
 from . import INFO, WARNING, LintCheck
 
 # Does not include m2-tools, which should be checked using wild cards.
@@ -36,7 +39,6 @@ PYTHON_BUILD_TOOLS = (
     "hatchling",
     "meson",
     "meson-python",
-    "packaging",
     "pdm",
     "pdm-pep517",
     "pip",
@@ -49,6 +51,25 @@ PYTHON_BUILD_TOOLS = (
     "whey",
 )
 
+# List of known PEP-517 backends (https://peps.python.org/pep-0517/) that are not setuptools
+# and that don't require wheel since they create wheels themselves.
+# The backend is defined using the pyproject.toml file (which was introduced by PEP-518).
+# Historical note: pyproject.toml file was introduced initially to specify the build system (backend).
+# Only later was the ability to specify all the project metadata (name, version, dependencies, etc)
+# into it added, via PEP-621.
+PYTHON_BUILD_BACKENDS = (
+    "flit",  # Our packages are not supposed to depend on flit, but apparently they do, so we need to support it here.
+    "flit-core",  # Backend of flit.
+    "hatch",  # Same as flit, we should not depend on it. We should instead depend on hatchling, which is the backend.
+    "hatchling",  # Backend of hatch.
+    "meson-python",  # Backend that uses meson.
+    "pdm-backend",  # Backend of pdm (new). Not yet in our repo, but if it ever does, we'll be able to handle it.
+    "pdm-pep517",  # Deprecated backend of pdm but we still need to support it, see https://pypi.org/project/pdm-pep517
+    "poetry-core",  # Backend of poetry.
+    "scikit-build-core",  # Backend that uses cmake.
+    "whey",
+)
+
 COMPILERS = (
     "cgo",
     "cuda",
@@ -95,14 +116,15 @@ def recipe_has_patches(recipe):
 
 
 class host_section_needs_exact_pinnings(LintCheck):
-    """Packages in host must have exact version pinnings, except python build tools.
+    """Linked libraries host should have exact version pinnings.
+    Other dependencies are case by case.
 
     Specifically, comparison operators must not be used. The version numbers can be
-    speficified in a conda_build_config.yaml file.
+    specified in a conda_build_config.yaml file.
     """
 
     def check_recipe(self, recipe):
-        deps = recipe.get_deps_dict("host")
+        deps = _utils.get_deps_dict(recipe, "host")
         for package, dep in deps.items():
             if not self.is_exception(package) and not (
                 package in recipe.selector_dict and recipe.selector_dict[package]
@@ -111,7 +133,7 @@ class host_section_needs_exact_pinnings(LintCheck):
                     if constraint == "" or re.search("^[<>!]", constraint) is not None:
                         path = dep["paths"][c]
                         output = -1 if not path.startswith("outputs") else int(path.split("/")[1])
-                        self.message(section=path, output=output)
+                        self.message(section=path, severity=WARNING, output=output)
 
     @staticmethod
     def is_exception(package):
@@ -119,12 +141,60 @@ class host_section_needs_exact_pinnings(LintCheck):
             "python",
             "toml",
             "wheel",
+            "packaging",
             *PYTHON_BUILD_TOOLS,
         )
         # It doesn't make sense to pin the versions of hatch plugins if we're not pinning
-        # hatch. We could explictly enumerate the 15 odd plugins in PYTHON_BUILD_TOOLS, but
-        # this seemed lower maintainance
-        return (package in exceptions) or package.startswith("hatch-")
+        # hatch. We could explicitly enumerate the 15 odd plugins in PYTHON_BUILD_TOOLS, but
+        # this seemed lower maintenance
+        return (package in exceptions) or any(
+            [package.startswith(f"{pkg}-") for pkg in PYTHON_BUILD_TOOLS]
+        )
+
+
+class cbc_dep_in_run_missing_from_host(LintCheck):
+    """Run dependencies listed in the cbc should also be present in the host section."""
+
+    def check_recipe(self, recipe):
+        for package in recipe.packages.values():
+            for dep in package.run:
+                if dep.pkg in recipe.selector_dict and recipe.selector_dict[dep.pkg]:
+                    if not self.is_exception(dep.pkg) and not package.has_dep("host", dep.pkg):
+                        dep_path = _utils.get_dep_path(recipe, dep)
+                        self.message(
+                            section=dep_path,
+                            data=(recipe, f"{package.path_prefix}requirements/host", dep.pkg),
+                        )
+
+    @staticmethod
+    def is_exception(package):
+        exceptions = (
+            "python",
+            "numpy",
+        )
+        return package in exceptions
+
+    def fix(self, _message, _data):
+        (recipe, path, dep) = _data
+        op = [
+            {
+                "op": "add",
+                "path": path,
+                "match": f"{dep}.*",
+                "value": [f"{dep} " + "{{ " + f"{dep}" + " }}"],
+            },
+        ]
+        return recipe.patch(op)
+
+
+class potentially_bad_ignore_run_exports(LintCheck):
+    """Ignoring run_export of a host dependency. In some cases it is more appropriate to remove the --error-overdepending flag of conda-build."""  # noqa: E501
+
+    def check_recipe(self, recipe):
+        for package in recipe.packages.values():
+            for dep in package.host:
+                if dep.pkg in package.ignore_run_exports:
+                    self.message(section=_utils.get_dep_path(recipe, dep), severity=INFO)
 
 
 class should_use_compilers(LintCheck):
@@ -196,7 +266,7 @@ class build_tools_must_be_in_build(LintCheck):
     """
 
     def check_recipe(self, recipe):
-        deps = recipe.get_deps_dict(["host", "run"])
+        deps = _utils.get_deps_dict(recipe, ["host", "run"])
         for tool, dep in deps.items():
             if tool.startswith("m2-") or tool in BUILD_TOOLS:
                 for path in dep["paths"]:
@@ -214,7 +284,7 @@ class python_build_tool_in_run(LintCheck):
     """
 
     def check_recipe(self, recipe):
-        deps = recipe.get_deps_dict("run")
+        deps = _utils.get_deps_dict(recipe, "run")
         for tool in PYTHON_BUILD_TOOLS:
             if tool in deps:
                 for path in deps[tool]["paths"]:
@@ -231,7 +301,7 @@ class missing_python_build_tool(LintCheck):
     def check_recipe(self, recipe):
         is_pypi = is_pypi_source(recipe)
         if outputs := recipe.get("outputs", None):
-            deps = recipe.get_deps_dict("host")
+            deps = _utils.get_deps_dict(recipe, "host")
             for o in range(len(outputs)):
                 # Create a list of build tool dependencies for each output
                 tools = []
@@ -240,12 +310,12 @@ class missing_python_build_tool(LintCheck):
                         path.startswith(f"outputs/{o}") for path in dep["paths"]
                     ):
                         tools.append(tool)
-                if (is_pypi or "pip install" in recipe.get(f"outputs/{o}/script", "")) and len(
+                if (is_pypi or recipe.contains(f"outputs/{o}/script", "pip install", "")) and len(
                     tools
                 ) == 0:
                     self.message(section=f"outputs/{o}/requirements/host", output=o)
-        elif is_pypi or "pip install" in recipe.get("build/script", ""):
-            deps = recipe.get_deps("host")
+        elif is_pypi or recipe.contains("build/script", "pip install", ""):
+            deps = _utils.get_deps(recipe, "host")
             if not any(tool in deps for tool in PYTHON_BUILD_TOOLS):
                 self.message(section="requirements/host")
 
@@ -261,22 +331,42 @@ class missing_wheel(LintCheck):
     """
 
     def check_recipe(self, recipe):
-        # similar algorithm as missing_python
         is_pypi = is_pypi_source(recipe)
-        deps = recipe.get_deps_dict("host")
-        paths = (
-            []
-            if "wheel" not in deps
-            else ["/".join(path.split("/")[:-1]) for path in deps["wheel"]["paths"]]
-        )
-        if outputs := recipe.get("outputs", None):
-            for o in range(len(outputs)):
-                if is_pypi or "pip install" in recipe.get(f"outputs/{o}/script", ""):
-                    if f"outputs/{o}/requirements/host" not in paths:
-                        self.message(section=f"outputs/{o}/requirements/host", output=o)
-        elif is_pypi or "pip install" in recipe.get("build/script", ""):
-            if "requirements/host" not in paths:
-                self.message(section="requirements/host")
+
+        for package in recipe.packages.values():
+            if (
+                is_pypi
+                or recipe.contains(f"{package.path_prefix}build/script", "pip install", "")
+                or recipe.contains(f"{package.path_prefix}script", "pip install", "")
+            ):
+                # Note that we do the assumption that if none of the backends defined in exceptions are present
+                # and that setuptools is also not present, that setuptools will be used via the good old
+                # setup.py file. This is because pip defaults to doing that for historical reasons.
+                # This means that if pip is used in the install script and there is no host dependencies,
+                # we want this check to raise a warning because wheel should be added there!
+                #
+                # In theory we would also need to warn if setuptools missing in the host
+                # section and none of the new build backends are used.
+                # TODO: add a missing_setuptools rule?
+                if not package.has_dep("host", "wheel") and not set(
+                    PYTHON_BUILD_BACKENDS
+                ).intersection({dep.pkg.lower() for dep in package.get("host")}):
+                    self.message(
+                        section=f"{package.path_prefix}requirements/host",
+                        data=(recipe, package),
+                    )
+
+    def fix(self, _message, _data):
+        (recipe, package) = _data
+        op = [
+            {
+                "op": "add",
+                "path": f"{package.path_prefix}requirements/host",
+                "match": "wheel",
+                "value": ["wheel"],
+            },
+        ]
+        return recipe.patch(op)
 
 
 class uses_setup_py(LintCheck):
@@ -290,42 +380,61 @@ class uses_setup_py(LintCheck):
     """
 
     @staticmethod
-    def _check_line(line: str) -> bool:
+    def _check_line(x: str) -> bool:
         """Check a line for a broken call to setup.py"""
-        if "setup.py install" in line:
-            return False
+        if isinstance(x, str):
+            x = [x]
+        elif not isinstance(x, list):
+            return True
+        for line in x:
+            if "setup.py install" in line:
+                return False
         return True
 
-    def check_deps(self, deps):
-        if "setuptools" not in deps:
-            return  # no setuptools, no problem
-
-        for path in deps["setuptools"]["paths"]:
-            if path.startswith("output"):
-                n = path.split("/")[1]
-                script = f"outputs/{n}/script"
-                output = int(n)
-            else:
-                script = "build/script"
-                output = -1
-            if not self._check_line(self.recipe.get(script, "")):
-                self.message(section=script)
-                continue
-            if self.recipe.dir:
+    def check_recipe(self, recipe):
+        for package in recipe.packages.values():
+            if not self._check_line(recipe.get(f"{package.path_prefix}build/script", None)):
+                self.message(
+                    section=f"{package.path_prefix}build/script",
+                    data=(recipe, f"{package.path_prefix}build/script"),
+                )
+            elif not self._check_line(recipe.get(f"{package.path_prefix}script", None)):
+                self.message(
+                    section=f"{package.path_prefix}script",
+                    data=(recipe, f"{package.path_prefix}script"),
+                )
+            elif self.recipe.dir:
                 try:
-                    if script == "build/script":
-                        build_file = "build.sh"
-                    else:
-                        build_file = self.recipe.get(script, "")
+                    build_file = self.recipe.get(f"{package.path_prefix}script", "")
                     if not build_file:
-                        continue
-                    with open(os.path.join(self.recipe.dir, build_file)) as buildsh:
-                        for num, line in enumerate(buildsh):
-                            if not self._check_line(line):
-                                self.message(fname=build_file, line=num, output=output)
+                        build_file = self.recipe.get(
+                            f"{package.path_prefix}build/script", "build.sh"
+                        )
+                    build_file = self.recipe.dir / Path(build_file)
+                    if build_file.exists():
+                        with open(str(build_file)) as buildsh:
+                            for num, line in enumerate(buildsh):
+                                if not self._check_line(line):
+                                    if package.path_prefix.startswith("output"):
+                                        output = int(package.path_prefix.split("/")[1])
+                                    else:
+                                        output = -1
+                                    self.message(fname=build_file, line=num, output=output)
                 except FileNotFoundError:
                     pass
 
+    def fix(self, _message, _data):
+        (recipe, path) = _data
+        op = [
+            {
+                "op": "replace",
+                "path": path,
+                "match": ".* setup.py .*",
+                "value": "{{PYTHON}} -m pip install . --no-deps --no-build-isolation --ignore-installed --no-cache-dir -vv",  # noqa: E501
+            },
+        ]
+        return recipe.patch(op)
+
 
 class pip_install_args(LintCheck):
     """`pip install` should be run with --no-deps and --no-build-isolation.
@@ -337,47 +446,65 @@ class pip_install_args(LintCheck):
     """
 
     @staticmethod
-    def _check_line(line: str) -> bool:
-        """Check a line for a broken call to setup.py"""
-        if "pip install" not in line:
+    def _check_line(x: Any) -> bool:
+        """Check a line (or list of lines) for a broken call to setup.py"""
+        if isinstance(x, str):
+            x = [x]
+        elif not isinstance(x, list):
             return True
 
-        required_args = ["--no-deps", "--no-build-isolation"]
-        if any(arg not in line for arg in required_args):
-            return False
+        for line in x:
+            if "pip install" in line:
+                required_args = ["--no-deps", "--no-build-isolation"]
+                if any(arg not in line for arg in required_args):
+                    return False
 
         return True
 
-    def check_deps(self, deps):
-        if "pip" not in deps:
-            return  # no pip, no problem
-
-        for path in deps["pip"]["paths"]:
-            if path.startswith("output"):
-                n = path.split("/")[1]
-                script = f"outputs/{n}/script"
-                output = int(n)
-            else:
-                script = "build/script"
-                output = -1
-            if not self._check_line(self.recipe.get(script, "")):
-                self.message(section=script)
-                continue
-            if self.recipe.dir:
+    def check_recipe(self, recipe):
+        for package in recipe.packages.values():
+            if not self._check_line(recipe.get(f"{package.path_prefix}build/script", None)):
+                self.message(
+                    section=f"{package.path_prefix}build/script",
+                    data=(recipe, f"{package.path_prefix}build/script"),
+                )
+            elif not self._check_line(recipe.get(f"{package.path_prefix}script", None)):
+                self.message(
+                    section=f"{package.path_prefix}script",
+                    data=(recipe, f"{package.path_prefix}script"),
+                )
+            elif self.recipe.dir:
                 try:
-                    if script == "build/script":
-                        build_file = "build.sh"
-                    else:
-                        build_file = self.recipe.get(script, "")
+                    build_file = self.recipe.get(f"{package.path_prefix}script", "")
                     if not build_file:
-                        continue
-                    with open(os.path.join(self.recipe.dir, build_file)) as buildsh:
-                        for num, line in enumerate(buildsh):
-                            if not self._check_line(line):
-                                self.message(fname=build_file, line=num, output=output)
+                        build_file = self.recipe.get(
+                            f"{package.path_prefix}build/script", "build.sh"
+                        )
+                    build_file = self.recipe.dir / Path(build_file)
+                    if build_file.exists():
+                        with open(str(build_file)) as buildsh:
+                            for num, line in enumerate(buildsh):
+                                if not self._check_line(line):
+                                    if package.path_prefix.startswith("output"):
+                                        output = int(package.path_prefix.split("/")[1])
+                                    else:
+                                        output = -1
+                                    self.message(fname=build_file, line=num, output=output)
                 except FileNotFoundError:
                     pass
 
+    def fix(self, _message, _data):
+        (recipe, path) = _data
+        op = [
+            {
+                "op": "replace",
+                "path": path,
+                "match": r"(.*\s)?pip install(?!=.*--no-build-isolation).*",
+                "value": "{{ PYTHON }} -m pip install . --no-deps --no-build-isolation --ignore-installed --no-cache-dir -vv",  # noqa: E501
+            },
+        ]
+        return recipe.patch(op)
+
 
 class cython_must_be_in_host(LintCheck):
     """Cython should be in the host section
@@ -448,14 +575,68 @@ class avoid_noarch(LintCheck):
     """
 
     def check_recipe(self, recipe):
-        noarch = recipe.get("build/noarch", "")
-        if (
-            noarch == "python"
-            and recipe.get("build/number", 0) == 0
-            and not recipe.get("build/osx_is_app", False)
-            and not recipe.get("app", None)
-        ):
-            self.message(section="build", severity=WARNING)
+        for package in recipe.packages.values():
+            noarch = recipe.get(f"{package.path_prefix}build/noarch", "")
+            if (
+                noarch == "python"
+                and int(recipe.get(f"{package.path_prefix}build/number", 0)) == 0
+                and not recipe.get(f"{package.path_prefix}build/osx_is_app", False)
+                and not recipe.get(f"{package.path_prefix}app", None)
+            ):
+                self.message(
+                    section=f"{package.path_prefix}build", severity=WARNING, data=(recipe, package)
+                )
+
+    def fix(self, _message, _data):
+        (recipe, package) = _data
+        skip_selector = None
+        sep_map = {
+            ">=": "<",
+            ">": "<=",
+            "==": "!=",
+            "!=": "==",
+            "<=": ">",
+            "<": ">=",
+        }
+        for dep in recipe.get(f"{package.path_prefix}requirements/run", []):
+            if dep.startswith("python"):
+                for sep, opp in sep_map.items():
+                    s = dep.split(sep)
+                    if len(s) > 1:
+                        skip_selector = f" # [py{opp}{s[1].strip().replace('.','')}]"
+                        break
+                if skip_selector:
+                    break
+        op = [
+            {"op": "remove", "path": f"{package.path_prefix}build/noarch"},
+            {
+                "op": "add",
+                "path": f"{package.path_prefix}requirements/host",
+                "match": "python",
+                "value": ["python"],
+            },
+            {
+                "op": "add",
+                "path": f"{package.path_prefix}requirements/run",
+                "match": "python",
+                "value": ["python"],
+            },
+            {
+                "op": "replace",
+                "path": f"{package.path_prefix}test/requires",
+                "match": "python",
+                "value": ["python"],
+            },
+        ]
+        if skip_selector:
+            op.append(
+                {
+                    "op": "add",
+                    "path": f"{package.path_prefix}build/skip",
+                    "value": f"True {skip_selector}",
+                }
+            )
+        return recipe.patch(op)
 
 
 class patch_unnecessary(LintCheck):
@@ -466,7 +647,9 @@ class patch_unnecessary(LintCheck):
 
     def check_recipe(self, recipe):
         if not recipe_has_patches(recipe):
-            deps = recipe.get_deps_dict()
+            deps = _utils.get_deps_dict(
+                recipe,
+            )
             if "patch" in deps or "m2-patch" in deps:
                 self.message(section="build")
 
@@ -484,7 +667,9 @@ class patch_must_be_in_build(LintCheck):
 
     def check_recipe(self, recipe):
         if recipe_has_patches(recipe):
-            deps = recipe.get_deps_dict()
+            deps = _utils.get_deps_dict(
+                recipe,
+            )
             if "patch" in deps:
                 if any("build" not in location for location in deps["patch"]["paths"]):
                     self.message(section="requirements/build")
@@ -539,7 +724,7 @@ class missing_imports_or_run_test_py(LintCheck):
 
     def check_recipe(self, recipe):
         is_pypi = is_pypi_source(recipe)
-        deps = recipe.get_deps_dict("host")
+        deps = _utils.get_deps_dict(recipe, "host")
         if not is_pypi and "python" not in deps:
             return
         if outputs := recipe.get("outputs", None):
@@ -547,7 +732,11 @@ class missing_imports_or_run_test_py(LintCheck):
             if is_pypi:
                 paths_to_check = [f"outputs/{o}" for o in range(len(outputs))]
             else:
-                paths_to_check = ["/".join(path.split("/")[:2]) for path in deps["python"]["paths"]]
+                paths_to_check = [
+                    "/".join(path.split("/")[:2])
+                    for path in deps["python"]["paths"]
+                    if not path.startswith("requirements")
+                ]
             for path in paths_to_check:
                 if not recipe.get(f"{path}/test/imports", []) and not recipe.get(
                     f"{path}/test/script", ""
@@ -572,7 +761,7 @@ class missing_pip_check(LintCheck):
           - pip check
     """
 
-    def _check_file(self, file, output=-1):
+    def _check_file(self, file, recipe, package):
         """Reads the file for a `pip check` command and outputs a message if not found.
 
         Args:
@@ -584,26 +773,37 @@ class missing_pip_check(LintCheck):
                 for line in test_file:
                     if "pip check" in line:
                         return
-        self.message(fname=file, output=output)
+        self.message(fname=file, data=(recipe, package))
 
     def check_recipe(self, recipe):
         is_pypi = is_pypi_source(recipe)
-        if outputs := recipe.get("outputs", None):
-            for o in range(len(outputs)):
-                if not is_pypi and "pip install" not in recipe.get(f"outputs/{o}/script", ""):
-                    continue
-                test_section = f"outputs/{o}/test"
-                if commands := recipe.get(f"{test_section}/commands", None):
-                    if not any("pip check" in cmd for cmd in commands):
-                        self.message(section=f"{test_section}/commands", output=o)
-                elif recipe.dir and (script := recipe.get(f"{test_section}/script", None)):
-                    self._check_file(os.path.join(recipe.dir, script))
-                else:
-                    self.message(section=test_section, output=o)
-        elif is_pypi or "pip install" in recipe.get("build/script", ""):
-            if commands := recipe.get("test/commands", None):
+
+        single_output = True
+        if len(recipe.packages) > 1:
+            single_output = False
+
+        for package in recipe.packages.values():
+            if (
+                not single_output
+                and not is_pypi
+                and not recipe.contains(f"{package.path_prefix}script", "pip install", "")
+            ):
+                continue
+            if (
+                single_output
+                and not is_pypi
+                and not recipe.contains(f"{package.path_prefix}build/script", "pip install", "")
+            ):
+                continue
+            if commands := recipe.get(f"{package.path_prefix}test/commands", None):
                 if not any("pip check" in cmd for cmd in commands):
-                    self.message(section="test/commands")
+                    self.message(
+                        section=f"{package.path_prefix}test/commands", data=(recipe, package)
+                    )
+            elif recipe.dir and (script := recipe.get(f"{package.path_prefix}test/script", None)):
+                self._check_file(os.path.join(recipe.dir, script), recipe, package)
+            elif not single_output:
+                self.message(section=f"{package.path_prefix}test", data=(recipe, package))
             else:
                 test_files = (
                     set(os.listdir(recipe.dir)).intersection({"run_test.sh", "run_test.bat"})
@@ -612,9 +812,21 @@ class missing_pip_check(LintCheck):
                 )
                 if len(test_files) > 0:
                     for file in test_files:
-                        self._check_file(os.path.join(recipe.dir, file))
+                        self._check_file(os.path.join(recipe.dir, file), recipe, package)
                 else:
-                    self.message(section="test")
+                    self.message(section=f"{package.path_prefix}test")
+
+    def fix(self, _message, _data):
+        (recipe, package) = _data
+        op = [
+            {
+                "op": "add",
+                "path": f"{package.path_prefix}test/commands",
+                "match": "pip check",
+                "value": ["pip check"],
+            },
+        ]
+        return recipe.patch(op)
 
 
 class missing_test_requirement_pip(LintCheck):
@@ -656,14 +868,23 @@ class missing_test_requirement_pip(LintCheck):
         return False
 
     def check_recipe(self, recipe):
-        if outputs := recipe.get("outputs", None):
-            for o in range(len(outputs)):
-                if self._has_pip_check(recipe, output=f"outputs/{o}/") and "pip" not in recipe.get(
-                    f"outputs/{o}/test/requires", []
-                ):
-                    self.message(section=f"outputs/{o}/test/requires/", output=o)
-        elif self._has_pip_check(recipe) and "pip" not in recipe.get("test/requires", []):
-            self.message(section="test/requires")
+        for package in recipe.packages.values():
+            if self._has_pip_check(recipe, output=package.path_prefix) and "pip" not in recipe.get(
+                f"{package.path_prefix}test/requires", []
+            ):
+                self.message(section=f"{package.path_prefix}test/requires", data=(recipe, package))
+
+    def fix(self, _message, _data):
+        (recipe, package) = _data
+        op = [
+            {
+                "op": "add",
+                "path": f"{package.path_prefix}test/requires",
+                "match": "pip",
+                "value": ["pip"],
+            },
+        ]
+        return recipe.patch(op)
 
 
 class missing_python(LintCheck):
@@ -679,31 +900,37 @@ class missing_python(LintCheck):
     """
 
     def check_recipe(self, recipe):
-        # To check for a missing python, find the paths that require python
-        # and compare with the existing dependency dictionary.
-        # Since this dictionary has everything parsed already, there is no need for regex.
         is_pypi = is_pypi_source(recipe)
-        deps = recipe.get_deps_dict(["host", "run"])
-        # The `paths` dictionary stores dependencies as, e.g., `requirements/host/{n}/`
-        # with the list index n. For multi-output recipes, it the paths are of the form
-        # `outputs/{o}/requirements/host/{n}`. To compare, the list index needs to be stripped.
-        paths = (
-            []
-            if "python" not in deps
-            else ["/".join(path.split("/")[:-1]) for path in deps["python"]["paths"]]
-        )
-        if outputs := recipe.get("outputs", None):
-            for o in range(len(outputs)):
-                if is_pypi or "pip install" in recipe.get(f"outputs/{o}/script", ""):
-                    for section in ["host", "run"]:
-                        path_to_section = f"outputs/{o}/requirements/{section}"
-                        if path_to_section not in paths:
-                            self.message(section, section=path_to_section, output=o)
-        elif is_pypi or "pip install" in recipe.get("build/script", ""):
-            for section in ["host", "run"]:
-                path_to_section = f"requirements/{section}"
-                if path_to_section not in paths:
-                    self.message(section, section=path_to_section)
+        for package in recipe.packages.values():
+            if (
+                is_pypi
+                or recipe.contains(f"{package.path_prefix}build/script", "pip install", "")
+                or recipe.contains(f"{package.path_prefix}script", "pip install", "")
+            ):
+                for section in ["host", "run"]:
+                    if not package.has_dep(section, "python"):
+                        self.message(
+                            section=f"{package.path_prefix}requirements/{section}",
+                            data=(recipe, package),
+                        )
+
+    def fix(self, _message, _data):
+        (recipe, package) = _data
+        op = [
+            {
+                "op": "add",
+                "path": f"{package.path_prefix}requirements/host",
+                "match": "python",
+                "value": ["python"],
+            },
+            {
+                "op": "add",
+                "path": f"{package.path_prefix}requirements/run",
+                "match": "python",
+                "value": ["python"],
+            },
+        ]
+        return recipe.patch(op)
 
 
 class remove_python_pinning(LintCheck):
@@ -722,15 +949,27 @@ class remove_python_pinning(LintCheck):
     """
 
     def check_recipe(self, recipe):
-        if recipe.get("build/noarch", "") == "":
-            sections = ["host", "run"]
-            deps = recipe.get_deps_dict(sections=sections)
-            if "python" in deps:
-                for c, constraint in enumerate(deps["python"]["constraints"]):
-                    if constraint != "":
-                        path = deps["python"]["paths"][c]
-                        output = -1 if not path.startswith("outputs") else int(path.split("/")[1])
-                        self.message(section=path, output=output)
+        for package in recipe.packages.values():
+            if recipe.get(f"{package.path_prefix}build/noarch", "") == "":
+                for section in ["host", "run"]:
+                    for dep in package.get(section):
+                        if dep.pkg == "python" and dep.constraint != "":
+                            self.message(section=dep.path, data=(recipe, package, dep))
+
+    def fix(self, _message, _data):
+        (recipe, package, dep) = _data
+        s = dep.split(">=")
+        if len(s) > 1:
+            skip_selector = f" # [py<{s[1].replace('.','')}]"
+            op = [
+                {
+                    "op": "add",
+                    "path": f"{package.path_prefix}build/skip",
+                    "value": f"True {skip_selector}",
+                }
+            ]
+            return recipe.patch(op)
+        return False
 
 
 class no_git_on_windows(LintCheck):
@@ -740,7 +979,7 @@ class no_git_on_windows(LintCheck):
     """
 
     def check_deps(self, deps):
-        if self.recipe.selector_dict["win"] == 1 and "git" in deps:
+        if self.recipe.selector_dict.get("win", 0) == 1 and "git" in deps:
             for path in deps["git"]["paths"]:
                 output = -1 if not path.startswith("outputs") else int(path.split("/")[1])
                 self.message(section=path, output=output)
@@ -765,5 +1004,5 @@ class gui_app(LintCheck):
     )
 
     def check_recipe(self, recipe):
-        if set(self.guis).intersection(set(recipe.get_deps("run"))):
+        if set(self.guis).intersection(set(_utils.get_deps(recipe, "run"))):
             self.message(severity=INFO)
